import unittest
from boarml.cnn.layers import *


class TestLayers(unittest.TestCase):

    def test_input_layer(self):
        # Given
        layer = InputLayer(shape=(3, 4))
        expected = 'InputLayer(shape=(3, 4))'

        # When
        actual = str(layer)

        # Then
        self.assertEqual(Layers.INPUT, layer.type)
        self.assertEqual(expected, actual)

    def test_conv_layer(self):
        # Given
        layer = ConvLayer(filters=64, kernel_size=(3, 3), strides=(2, 2))
        expected = 'ConvLayer(filters=64; kernel_size=(3, 3); strides=(2, 2))'

        # When
        actual = str(layer)

        # Then
        self.assertEqual(Layers.CONV, layer.type)
        self.assertEqual(expected, actual)

    def test_activation_layer(self):
        # Given
        layer = ActivationLayer(func='relu')
        expected = 'ActivationLayer(func="relu")'

        # When
        actual = str(layer)

        # Then
        self.assertEqual(Layers.ACTIVATION, layer.type)
        self.assertEqual(expected, actual)

    def test_dense_layer(self):
        # Given
        layer = DenseLayer(nodes=10)
        expected = 'DenseLayer(nodes=10)'

        # When
        actual = str(layer)

        # Then
        self.assertEqual(Layers.DENSE, layer.type)
        self.assertEqual(expected, actual)

    def test_flatten_layer(self):
        # Given
        layer = FlattenLayer()
        expected = 'FlattenLayer()'

        # When
        actual = str(layer)

        # Then
        self.assertEqual(Layers.FLATTEN, layer.type)
        self.assertEqual(expected, actual)

    def test_batch_normalization_layer(self):
        # Given
        layer = BatchNormalizationLayer(momentum=0.99)
        expected = 'BatchNormalizationLayer(momentum=0.99)'

        # When
        actual = str(layer)

        # Then
        self.assertEqual(Layers.BATCH_NORM, layer.type)
        self.assertEqual(expected, actual)

    def test_pooling_layer(self):
        # Given
        layer = PoolingLayer(operation='max', pool_size=(2, 2), strides=(2, 2))
        expected = 'PoolingLayer(operation="max"; pool_size=(2, 2); strides=(2, 2))'

        # When
        actual = str(layer)

        # Then
        self.assertEqual(Layers.POOLING, layer.type)
        self.assertEqual(expected, actual)

    def test_droput_layer(self):
        # Given
        layer = DropoutLayer(rate=0.2)
        expected = 'DropoutLayer(rate=0.2)'

        # When
        actual = str(layer)

        # Then
        self.assertEqual(Layers.DROPOUT, layer.type)
        self.assertEqual(expected, actual)



from boarml.cnn import *
from enum import IntEnum


class ArchSection(IntEnum):
    INPUT = 1
    MAIN = 2
    OUTPUT = 3
    START = 4
    END = 5
    UNIT = 6


MAIN_ARCH_SYMBOLS = {ArchSection.START: '<<<<<START>>>>>', ArchSection.INPUT: '---Input---', ArchSection.MAIN: '---Main---',
                     ArchSection.UNIT: Unit.UNIT_SYMBOL,
                     ArchSection.OUTPUT: '---Output---', ArchSection.END: '<<<<<END>>>>>'}


class CnnBaseArchitecture:

    def __init__(self, input_shape=0, output_nodes=0):
        self.layers = dict()

        self.input_shape = input_shape
        self.output_shape = output_nodes

        self.__add_input_layers(input_shape)
        self.__add_output_layers(output_nodes)

        self.layers[ArchSection.MAIN] = list({Unit()})
        self.main_units_count = 0

    def __add_input_layers(self, input_shape):
        input_layers = list()
        input_layers.append(InputLayer(shape=input_shape))
        self.layers[ArchSection.INPUT] = input_layers

    def __add_output_layers(self, output_shape):
        output_layers = list()
        output_layers.append(FlattenLayer())
        output_layers.append(DenseLayer(nodes=output_shape))
        output_layers.append(ActivationLayer('softmax'))

        self.layers[ArchSection.OUTPUT] = output_layers

    def new_unit(self):
        if self.layers[ArchSection.MAIN][self.main_units_count].size > 0:
            self.layers[ArchSection.MAIN].append(Unit())
            self.main_units_count = self.main_units_count + 1

    def add_to_unit(self, layer):
        self.layers[ArchSection.MAIN][self.main_units_count].add_layer(layer)

    @property
    def input_arch(self):
        return self.layers[ArchSection.INPUT]

    @input_arch.setter
    def input_arch(self, input_layers):
        self.layers[ArchSection.INPUT] = input_layers

    @property
    def main_arch(self):
        return self.layers[ArchSection.MAIN]

    @main_arch.setter
    def main_arch(self, units):
        self.layers[ArchSection.MAIN] = units

    @property
    def output_arch(self):
        return self.layers[ArchSection.OUTPUT]

    @output_arch.setter
    def output_arch(self, output_layers):
        self.layers[ArchSection.OUTPUT] = output_layers

    def __str__(self):
        text = ''
        newline = '\n'

        text = f'{text}{MAIN_ARCH_SYMBOLS[ArchSection.START]}'

        # Convert Input to string
        input_count = len(self.input_arch)
        text = f'{text}{newline}{MAIN_ARCH_SYMBOLS[ArchSection.INPUT]}({input_count})'
        for layer in self.input_arch:
            text = f'{text}{newline}{layer}'

        # Convert Main to string
        main_count = len(self.main_arch)
        text = f'{text}{newline}{MAIN_ARCH_SYMBOLS[ArchSection.MAIN]}({main_count})'
        for main_unit in self.main_arch:
            text = f'{text}{newline}{main_unit}'

        # Convert Output to string
        output_count = len(self.output_arch)
        text = f'{text}{newline}{MAIN_ARCH_SYMBOLS[ArchSection.OUTPUT]}({output_count})'
        for layer in self.output_arch:
            text = f'{text}{newline}{layer}'

        text = f'{text}{newline}{MAIN_ARCH_SYMBOLS[ArchSection.END]}'

        return text

    def build_from_file(self, path):
        with open(path, 'r+') as file:
            text = file.read().splitlines()
            return self.build_from_text(text)

    def build_from_text(self, text):

        if text[0] != MAIN_ARCH_SYMBOLS[ArchSection.START]:
            raise Exception(
                f'The start of the architecture is not well defined!\nYou must use {MAIN_ARCH_SYMBOLS[ArchSection.START]} to indicate the start!')

        if text[-1] != MAIN_ARCH_SYMBOLS[ArchSection.END]:
            raise Exception(
                f'The end of the architecture is not well defined!\nYou must use {MAIN_ARCH_SYMBOLS[ArchSection.END]} to indicate the end!')

        line_index = 1

        if text[line_index].startswith(MAIN_ARCH_SYMBOLS[ArchSection.INPUT]):
            line_index = self.build_input_from_text(line_index, text)

        if text[line_index].startswith(MAIN_ARCH_SYMBOLS[ArchSection.MAIN]):
            line_index = self.build_main_from_text(line_index, text)

        if text[line_index].startswith(MAIN_ARCH_SYMBOLS[ArchSection.OUTPUT]):
            self.build_output_from_text(line_index, text)

        return self

    def build_output_from_text(self, line_index, text):
        layers_count = self.extract_number(text[line_index])
        otput_layers = list()

        for layer_index in range(1, layers_count + 1):
            layer = LayerBuilder.build_from_text(text[line_index + layer_index])

            if layer.type == Layers.DENSE:
                self.output_shape = layer.nodes

            otput_layers.append(layer)

        self.output_arch = otput_layers

    def build_main_from_text(self, line_index, text):
        unit_count = self.extract_number(text[line_index])
        main_units = list()
        line_index = line_index + 1

        for unit_index in range(1, unit_count + 1):

            layers_count = self.extract_number(text[line_index])
            unit = Unit()

            for layer_index in range(1, layers_count + 1):
                unit.add_layer(LayerBuilder.build_from_text(text[line_index + layer_index]))

            main_units.append(unit)

            line_index = line_index + layers_count + 1
        self.main_arch = main_units
        return line_index

    def build_input_from_text(self, line_index, text):
        layers_count = self.extract_number(text[line_index])
        input_layers = list()

        for layer_index in range(1, layers_count + 1):

            layer = LayerBuilder.build_from_text(text[line_index + layer_index])

            if layer.type == Layers.INPUT:
                self.input_shape = layer.shape

            input_layers.append(layer)

        self.input_arch = input_layers
        line_index = line_index + layers_count + 1

        return line_index

    def extract_number(self, line):

        return int(''.join(filter(str.isdigit, line)))



import unittest
from boarml.cnn import *


class TestUnit(unittest.TestCase):

    def test_empty_unit(self):
        # Given
        test_unit = Unit()
        expected = '-Unit-(0)'

        # When
        actual = str(test_unit)

        # Then
        self.assertEqual(expected, actual)
        self.assertEqual(test_unit.size, 0)

    def test_single_layer_unit(self):
        # Given
        test_unit = Unit()

        with open('resources/single_layer_unit.txt', 'r') as input_file:
            expected = input_file.read()

        # When
        test_unit.add_layer(DenseLayer(5))
        actual = str(test_unit)

        # Then
        self.assertEqual(expected, actual)
        self.assertEqual(test_unit.size, 1)

    def test_multi_layer_unit(self):
        # Given
        test_unit = Unit()

        with open('resources/multi_layer_unit.txt', 'r') as input_file:
            expected = input_file.read()

        # When
        test_unit.add_layer(DenseLayer(nodes=5))
        test_unit.add_layer(FlattenLayer())
        test_unit.add_layer(BatchNormalizationLayer(momentum=3))
        test_unit.add_layer(ConvLayer(filters=96, kernel_size=(3, 3), strides=(1, 1)))

        actual = str(test_unit)

        # Then
        self.assertEqual(expected, actual)
        self.assertEqual(test_unit.size, 4)



import copy, os

from boarml.architectures.cnn_architecture import CnnBaseArchitecture
from boarml.cnn.layers import Layers, ConvLayer, DropoutLayer, BatchNormalizationLayer, PoolingLayer
from boarml.cnn.unit import Unit
from random import Random
from boarml.builders.keras_builder import KerasBuilder


class ModelGenerator:

    def __init__(self, architecture, builder, removal_rate=0.5, duplication_rate=0.5, amendment_rate=0.5, seed=None):
        self.architecture = architecture
        self.builder = builder
        self.removal_rate = removal_rate
        self.duplication_rate = duplication_rate
        self.amendment_rate = amendment_rate
        self.rnd = Random(seed)

        self.mutation_history = list()

    def create_base_model(self):
        return self.create_model(self.architecture)

    def create_mutated_model(self):

        generate_new = True

        while generate_new:
            generate_new = False

            main_units = self.mutate_main_units(self.architecture)
            mutated_architecture = CnnBaseArchitecture(self.architecture.input_shape, self.architecture.output_shape)
            mutated_architecture.main_arch = main_units

            # If is in the history generate new
            for m in self.mutation_history:
                if str(m) == str(mutated_architecture):
                    generate_new = True

            # If it is invalid generate new
            if not self.is_valid_architecture(mutated_architecture):
                generate_new = True

        self.mutation_history.append(mutated_architecture)
        return self.create_model(mutated_architecture)

    def create_model(self, architecture):

        layers = list()

        for layer in architecture.input_arch:
            layers.append(layer)

        for unit in architecture.main_arch:
            for layer in unit.layers:
                layers.append(layer)

        for layer in architecture.output_arch:
            layers.append(layer)

        if self.builder == 'keras':
            model = KerasBuilder().build(layers)
            return model
        else:
            raise Exception('Unknown builder!')

    def mutate_main_units(self, architecture):

        units = list()

        for unit in architecture.main_arch:

            # Duplicate
            if self.duplication_rate > self.rnd.random():
                units.append(copy.deepcopy(unit))

            # Remove
            if self.removal_rate > self.rnd.random():
                pass
            else:
                # Amend
                if self.amendment_rate > self.rnd.random():
                    units.append(self.unit_mutation(unit))
                else:
                    units.append(unit)

        return units

    def unit_mutation(self, unit):
        mutated_unit = Unit()

        for layer in unit.layers:

            # Mutate Convolutional Layer
            if layer.type == Layers.CONV:
                mutated_layer = self.mutate_conv(layer)
            # Mutate Batch Normalisation Layer
            elif layer.type == Layers.BATCH_NORM:
                mutated_layer = self.mutate_batch_norm(layer)
            # Mutate Max Pooling Layer
            elif layer.type == Layers.POOLING:
                mutated_layer = self.mutate_pooling(layer)
            # Mutate Dropout Layer
            elif layer.type == Layers.DROPOUT:
                mutated_layer = self.mutate_dropout(layer)
            # Just add the layer
            else:
                mutated_layer = layer

            if mutated_layer is not None:
                mutated_unit.add_layer(mutated_layer)

        return mutated_unit

    def mutate_conv(self, layer):

        mutated = None

        choices = ['filter', 'kernel', 'stride']

        selected = self.rnd.choice(choices)

        if selected == 'filter':
            filters_choices = [64, 128, 256, 320, 512, 700]
            filter_selected = self.rnd.choice(filters_choices)

            mutated = ConvLayer(filters=filter_selected, kernel_size=layer.kernel_size, strides=layer.strides)

        if selected == 'kernel':
            kernel_choice = [(1, 1), (3, 3), (5, 5), (7, 7)]
            kernel_selected = self.rnd.choice(kernel_choice)

            mutated = ConvLayer(filters=layer.filters, kernel_size=kernel_selected, strides=layer.strides)

        if selected == 'stride':
            stride_choice = [(1, 1), (2, 2), (3, 3)]
            stride_selected = self.rnd.choice(stride_choice)

            mutated = ConvLayer(filters=layer.filters, kernel_size=layer.kernel_size, strides=stride_selected)

        return mutated

    def mutate_dropout(self, layer):

        mutated = None

        choices = ['rate', 'remove']

        selected = self.rnd.choice(choices)

        if selected == 'rate':
            rate_choices = [0.1, 0.2, 0.3, 0.4, 0.5]
            rate_selected = self.rnd.choice(rate_choices)

            mutated = DropoutLayer(rate=rate_selected)

        if selected == 'remove':
            pass

        return mutated

    def mutate_batch_norm(self, layer):

        mutated = None

        choices = ['momentum', 'remove']

        selected = self.rnd.choice(choices)

        if selected == 'momentum':
            momentum_choices = [0.99, 0.95, 0.90, 0.85]
            momentum_selected = self.rnd.choice(momentum_choices)

            mutated = BatchNormalizationLayer(momentum=momentum_selected)

        if selected == 'remove':
            pass

        return mutated

    def mutate_pooling(self, layer):

        mutated = None

        choices = ['operation', 'pool_size', 'remove']

        selected = self.rnd.choice(choices)

        if selected == 'operation':
            operation_choices = ['max', 'average']
            operation_selected = self.rnd.choice(operation_choices)

            mutated = PoolingLayer(operation=operation_selected, pool_size=layer.pool_size, strides=layer.strides)

        if selected == 'pool_size':
            pool_size_choices = [(2, 2), (3, 3)]
            pool_size_selected = self.rnd.choice(pool_size_choices)

            mutated = PoolingLayer(operation=layer.operation, pool_size=pool_size_selected, strides=layer.strides)

        if selected == 'remove':
            pass

        return mutated

    def get_last_mutation(self):
        if len(self.mutation_history) > 0:
            return self.mutation_history[-1]
        else:
            return self.architecture

    def populate_history(self, path, name, count):

        for i in range(0, count):
            full_name = name + str(i) + '.txt'

            full_path = os.path.join(path, full_name)
            arch = CnnBaseArchitecture()
            arch.build_from_file(full_path)

            self.mutation_history.append(arch)

    def is_valid_architecture(self, architecture):
        valid = True

        if len(architecture.main_arch) == 0:
            valid = False

        return valid



from enum import IntEnum
from abc import ABC, abstractmethod
from ast import literal_eval as make_tuple


class Layers(IntEnum):
    INPUT = 1
    CONV = 2
    ACTIVATION = 3
    DENSE = 4
    FLATTEN = 5
    BATCH_NORM = 6
    POOLING = 7
    DROPOUT = 8


MAIN_LAYER_TYPES = {Layers.INPUT: 'InputLayer', Layers.CONV: 'ConvLayer', Layers.ACTIVATION: 'ActivationLayer', Layers.DENSE: 'DenseLayer',
                    Layers.FLATTEN: 'FlattenLayer',
                    Layers.BATCH_NORM: 'BatchNormalizationLayer', Layers.POOLING: 'PoolingLayer', Layers.DROPOUT: 'DropoutLayer'}


class Layer(ABC):

    @property
    @abstractmethod
    def type(self):
        pass

    @abstractmethod
    def __str__(self):
        pass

    def convert_to_tuple(self, parameter):

        if isinstance(parameter, str):
            return make_tuple(parameter)
        else:
            return parameter

    def convert_to_int(self, parameter):

        if isinstance(parameter, str):
            return int(parameter)
        else:
            return parameter

    def convert_to_float(self, parameter):

        if isinstance(parameter, str):
            return float(parameter)
        else:
            return parameter


class InputLayer(Layer):

    def __init__(self, shape):
        self.shape = self.convert_to_tuple(shape)

    @property
    def type(self):
        return Layers.INPUT

    def __str__(self):
        return f'{MAIN_LAYER_TYPES[self.type]}(shape={self.shape})'


class ConvLayer(Layer):

    def __init__(self, filters, kernel_size, strides):
        self.filters = self.convert_to_int(filters)
        self.kernel_size = self.convert_to_tuple(kernel_size)
        self.strides = self.convert_to_tuple(strides)

    @property
    def type(self):
        return Layers.CONV

    def __str__(self):
        return f'{MAIN_LAYER_TYPES[self.type]}(filters={self.filters}; kernel_size={self.kernel_size}; strides={self.strides})'


class ActivationLayer(Layer):

    def __init__(self, func):
        self.func = func

    @property
    def type(self):
        return Layers.ACTIVATION

    def __str__(self):
        return f'{MAIN_LAYER_TYPES[self.type]}(func="{self.func}")'


class DenseLayer(Layer):

    def __init__(self, nodes):
        self.nodes = self.convert_to_int(nodes)

    @property
    def type(self):
        return Layers.DENSE

    def __str__(self):
        return f'{MAIN_LAYER_TYPES[self.type]}(nodes={self.nodes})'


class FlattenLayer(Layer):

    @property
    def type(self):
        return Layers.FLATTEN

    def __str__(self):
        return f'{MAIN_LAYER_TYPES[self.type]}()'


class BatchNormalizationLayer(Layer):

    def __init__(self, momentum):
        self.momentum = self.convert_to_float(momentum)

    @property
    def type(self):
        return Layers.BATCH_NORM

    def __str__(self):
        return f'{MAIN_LAYER_TYPES[self.type]}(momentum={self.momentum})'


class PoolingLayer(Layer):

    def __init__(self, operation, pool_size, strides):
        self.operation = operation
        self.pool_size = self.convert_to_tuple(pool_size)
        self.strides = self.convert_to_tuple(strides)

    @property
    def type(self):
        return Layers.POOLING

    def __str__(self):
        return f'{MAIN_LAYER_TYPES[self.type]}(operation="{self.operation}"; pool_size={self.pool_size}; strides={self.strides})'


class DropoutLayer(Layer):

    def __init__(self, rate):
        self.rate = self.convert_to_float(rate)

    @property
    def type(self):
        return Layers.DROPOUT

    def __str__(self):
        return f'{MAIN_LAYER_TYPES[self.type]}(rate={self.rate})'


class LayerBuilder:

    @staticmethod
    def build_from_text(text):

        parameters = LayerBuilder.extract_parameters(text)

        if text.startswith(MAIN_LAYER_TYPES[Layers.INPUT]):
            return InputLayer(**parameters)
        if text.startswith(MAIN_LAYER_TYPES[Layers.CONV]):
            return ConvLayer(**parameters)
        if text.startswith(MAIN_LAYER_TYPES[Layers.ACTIVATION]):
            return ActivationLayer(**parameters)
        if text.startswith(MAIN_LAYER_TYPES[Layers.DENSE]):
            return DenseLayer(**parameters)
        if text.startswith(MAIN_LAYER_TYPES[Layers.FLATTEN]):
            return FlattenLayer(**parameters)
        if text.startswith(MAIN_LAYER_TYPES[Layers.BATCH_NORM]):
            return BatchNormalizationLayer(**parameters)
        if text.startswith(MAIN_LAYER_TYPES[Layers.POOLING]):
            return PoolingLayer(**parameters)
        if text.startswith(MAIN_LAYER_TYPES[Layers.DROPOUT]):
            return DropoutLayer(**parameters)

    @staticmethod
    def extract_parameters(text):

        params_dict = dict()
        parameters = text.split('(', maxsplit=1)[1][:-1]

        if not parameters:
            return params_dict

        parameters = parameters.split('; ')

        for p in parameters:
            key_value_list = p.split('=')
            params_dict[key_value_list[0].strip('\"')] = key_value_list[1].strip('\"')

        return params_dict



import unittest
from boarml.architectures import CnnBaseArchitecture
from boarml.cnn import *


class TestCnnArchitecture(unittest.TestCase):

    def test_empty_architecture(self):
        # Given
        architecture = CnnBaseArchitecture(input_shape=(32, 32), output_nodes=10)

        with open('resources/empty_architecture.txt', 'r') as input_file:
            expected = input_file.read()

        # When
        actual = str(architecture)

        # Then
        self.assertEqual(expected, actual)

        self.assertEqual(Layers.INPUT, architecture.input_arch[0].type)
        self.assertEqual((32, 32), architecture.input_arch[0].shape)
        self.assertEqual(Layers.FLATTEN, architecture.output_arch[0].type)
        self.assertEqual(Layers.DENSE, architecture.output_arch[1].type)
        self.assertEqual(10, architecture.output_arch[1].nodes)
        self.assertEqual('softmax', architecture.output_arch[2].func)

    def test_simple_architecture(self):
        # Given
        architecture = CnnBaseArchitecture(input_shape=(32, 32), output_nodes=10)

        with open('resources/simple_architecture.txt', 'r') as input_file:
            expected = input_file.read()

        # When

        # Unit
        architecture.new_unit()
        architecture.add_to_unit(ConvLayer(filters=64, kernel_size=(3, 3), strides=(2, 2)))
        architecture.add_to_unit(BatchNormalizationLayer(momentum=0.999))
        architecture.add_to_unit(PoolingLayer(operation='max', pool_size=(2, 2), strides=(2, 2)))
        architecture.add_to_unit(DropoutLayer(rate=0.2))
        architecture.add_to_unit(ActivationLayer(func='relu'))

        actual = str(architecture)

        # Then
        self.assertEqual(expected, actual)

    def test_complex_architecture(self):
        # Given
        architecture = CnnBaseArchitecture(input_shape=(32, 32, 3), output_nodes=10)

        with open('resources/complex_architecture.txt', 'r') as input_file:
            expected = input_file.read()

        # When

        # Unit 1
        architecture.new_unit()
        architecture.add_to_unit(ConvLayer(filters=64, kernel_size=(3, 3), strides=(2, 2)))
        architecture.add_to_unit(BatchNormalizationLayer(momentum=0.999))
        architecture.add_to_unit(PoolingLayer(operation='max', pool_size=(2, 2), strides=(2, 2)))
        architecture.add_to_unit(DropoutLayer(rate=0.2))
        architecture.add_to_unit(ActivationLayer(func='relu'))

        # Unit 2
        architecture.new_unit()
        architecture.add_to_unit(ConvLayer(filters=64, kernel_size=(4, 4), strides=(2, 2)))
        architecture.add_to_unit(BatchNormalizationLayer(momentum=0.999))
        architecture.add_to_unit(PoolingLayer(operation='max', pool_size=(2, 2), strides=(2, 2)))
        architecture.add_to_unit(DropoutLayer(rate=0.2))
        architecture.add_to_unit(ActivationLayer(func='relu'))

        # Unit 3
        architecture.new_unit()
        architecture.add_to_unit(ConvLayer(filters=64, kernel_size=(3, 3), strides=(2, 2)))
        architecture.add_to_unit(BatchNormalizationLayer(momentum=0.999))
        architecture.add_to_unit(PoolingLayer(operation='average', pool_size=(2, 2), strides=(2, 2)))
        architecture.add_to_unit(DropoutLayer(rate=0.2))
        architecture.add_to_unit(ActivationLayer(func='relu'))

        actual = str(architecture)

        # Then
        self.assertEqual(expected, actual)

    def test_custom_input_layers_architecture(self):
        # Given
        architecture = CnnBaseArchitecture(input_shape=(32, 32), output_nodes=10)

        with open('resources/custom_input_layers_architecture.txt', 'r') as input_file:
            expected = input_file.read()

        # When
        input_layers = list()
        input_layers.append(InputLayer(shape=(13, 13)))
        input_layers.append(ConvLayer(filters=64, kernel_size=(4, 4), strides=(2, 2)))
        input_layers.append(ActivationLayer(func='relu'))
        architecture.input_arch = input_layers

        actual = str(architecture)

        # Then
        self.assertEqual(expected, actual)

        self.assertEqual(Layers.INPUT, architecture.input_arch[0].type)
        self.assertEqual((13, 13), architecture.input_arch[0].shape)
        self.assertEqual(Layers.CONV, architecture.input_arch[1].type)
        self.assertEqual(64, architecture.input_arch[1].filters)
        self.assertEqual((4, 4), architecture.input_arch[1].kernel_size)
        self.assertEqual((2, 2), architecture.input_arch[1].strides)
        self.assertEqual(Layers.ACTIVATION, architecture.input_arch[2].type)
        self.assertEqual('relu', architecture.input_arch[2].func)

    def test_custom_output_layers_architecture(self):
        # Given
        architecture = CnnBaseArchitecture(input_shape=(32, 32), output_nodes=10)

        with open('resources/custom_output_layers_architecture.txt', 'r') as input_file:
            expected = input_file.read()

        # When
        output_layers = list()
        output_layers.append(PoolingLayer(operation='max', pool_size=(2, 2), strides=(2, 2)))
        output_layers.append(DropoutLayer(rate=0.2))
        output_layers.append(FlattenLayer())
        architecture.output_arch = output_layers

        actual = str(architecture)

        # Then
        self.assertEqual(expected, actual)

        self.assertEqual(Layers.POOLING, architecture.output_arch[0].type)
        self.assertEqual('max', architecture.output_arch[0].operation)
        self.assertEqual((2, 2), architecture.output_arch[0].pool_size)
        self.assertEqual((2, 2), architecture.output_arch[0].strides)
        self.assertEqual(Layers.DROPOUT, architecture.output_arch[1].type)
        self.assertEqual(0.2, architecture.output_arch[1].rate)
        self.assertEqual(Layers.FLATTEN, architecture.output_arch[2].type)

    def test_custom_main_layers_architecture(self):
        # Given
        architecture = CnnBaseArchitecture(input_shape=(32, 32), output_nodes=10)

        with open('resources/custom_main_layers_architecture.txt', 'r') as input_file:
            expected = input_file.read()

        # When
        main_layers = list()

        # Unit 1
        new_unit = Unit()
        new_unit.add_layer(ConvLayer(filters=64, kernel_size=(3, 3), strides=(2, 2)))
        new_unit.add_layer(BatchNormalizationLayer(momentum=0.999))
        main_layers.append(new_unit)

        # Unit 2
        new_unit = Unit()
        new_unit.add_layer(PoolingLayer(operation='max', pool_size=(2, 2), strides=(2, 2)))
        new_unit.add_layer(DropoutLayer(rate=0.2))
        new_unit.add_layer(ActivationLayer(func='relu'))
        main_layers.append(new_unit)

        architecture.main_arch = main_layers

        actual = str(architecture)

        # Then
        self.assertEqual(expected, actual)

        self.assertEqual(Layers.CONV, architecture.main_arch[0].layers[0].type)
        self.assertEqual(64, architecture.main_arch[0].layers[0].filters)
        self.assertEqual((3, 3), architecture.main_arch[0].layers[0].kernel_size)
        self.assertEqual((2, 2), architecture.main_arch[0].layers[0].strides)

        self.assertEqual(Layers.BATCH_NORM, architecture.main_arch[0].layers[1].type)
        self.assertEqual(0.999, architecture.main_arch[0].layers[1].momentum)

        self.assertEqual(Layers.POOLING, architecture.main_arch[1].layers[0].type)
        self.assertEqual('max', architecture.main_arch[1].layers[0].operation)
        self.assertEqual((2, 2), architecture.main_arch[1].layers[0].pool_size)
        self.assertEqual((2, 2), architecture.main_arch[1].layers[0].strides)

        self.assertEqual(Layers.DROPOUT, architecture.main_arch[1].layers[1].type)
        self.assertEqual(0.2, architecture.main_arch[1].layers[1].rate)

        self.assertEqual(Layers.ACTIVATION, architecture.main_arch[1].layers[2].type)
        self.assertEqual('relu', architecture.main_arch[1].layers[2].func)

    def test_build_empty_from_text(self):
        # Given
        expected = CnnBaseArchitecture(input_shape=(32, 32), output_nodes=10)

        # When
        actual = CnnBaseArchitecture()
        actual = actual.build_from_file('resources/empty_architecture.txt')

        # Then
        self.assertEqual(str(expected), str(actual))

    def test_build_simple_from_text(self):
        # Given
        expected = CnnBaseArchitecture(input_shape=(32, 32), output_nodes=10)
        expected.new_unit()
        expected.add_to_unit(ConvLayer(filters=64, kernel_size=(3, 3), strides=(2, 2)))
        expected.add_to_unit(BatchNormalizationLayer(momentum=0.999))
        expected.add_to_unit(PoolingLayer(operation="max", pool_size=(2, 2), strides=(2, 2)))
        expected.add_to_unit(DropoutLayer(rate=0.2))
        expected.add_to_unit(ActivationLayer(func="relu"))
        expected.output_arch = [FlattenLayer(), DenseLayer(nodes=10), ActivationLayer(func="softmax")]

        # When
        actual = CnnBaseArchitecture()
        actual = actual.build_from_file('resources/simple_architecture.txt')

        # Then
        self.assertEqual(str(expected), str(actual))

    def test_build_complex_from_text(self):
        # Given
        expected = CnnBaseArchitecture(input_shape=(32, 32, 3), output_nodes=10)

        # Unit 1
        expected.new_unit()
        expected.add_to_unit(ConvLayer(filters=64, kernel_size=(3, 3), strides=(2, 2)))
        expected.add_to_unit(BatchNormalizationLayer(momentum=0.999))
        expected.add_to_unit(PoolingLayer(operation="max", pool_size=(2, 2), strides=(2, 2)))
        expected.add_to_unit(DropoutLayer(rate=0.2))
        expected.add_to_unit(ActivationLayer(func='relu'))

        # Unit 2
        expected.new_unit()
        expected.add_to_unit(ConvLayer(filters=64, kernel_size=(4, 4), strides=(2, 2)))
        expected.add_to_unit(BatchNormalizationLayer(momentum=0.999))
        expected.add_to_unit(PoolingLayer(operation="max", pool_size=(2, 2), strides=(2, 2)))
        expected.add_to_unit(DropoutLayer(rate=0.2))
        expected.add_to_unit(ActivationLayer(func='relu'))

        # Unit 3
        expected.new_unit()
        expected.add_to_unit(ConvLayer(filters=64, kernel_size=(3, 3), strides=(2, 2)))
        expected.add_to_unit(BatchNormalizationLayer(momentum=0.999))
        expected.add_to_unit(PoolingLayer(operation="average", pool_size=(2, 2), strides=(2, 2)))
        expected.add_to_unit(DropoutLayer(rate=0.2))
        expected.add_to_unit(ActivationLayer(func='relu'))

        expected.output_arch = [FlattenLayer(), DenseLayer(nodes=10), ActivationLayer(func='softmax')]

        # When
        actual = CnnBaseArchitecture()
        actual = actual.build_from_file('resources/complex_architecture.txt')

        # Then
        self.assertEqual(str(expected), str(actual))

    def test_build_invalid_start(self):
        # When
        actual = CnnBaseArchitecture()

        # Then
        self.assertRaises(Exception, actual.build_from_file, 'resources/invalid_start_architecture.txt')

    def test_build_invalid_end(self):
        # When
        actual = CnnBaseArchitecture()

        # Then
        self.assertRaises(Exception, actual.build_from_file, 'resources/invalid_end_architecture.txt')



import unittest
from boarml.generators import *
from boarml.architectures import *


class TestModelGenerator(unittest.TestCase):

    def test_create_base_model(self):
        # Given
        architecture = CnnBaseArchitecture()
        architecture.build_from_file('resources/complex_architecture.txt')

        # When
        generator = ModelGenerator(architecture, 'keras')
        actual = generator.create_base_model()

        # Then
        self.assertEqual(19, len(actual.model._layers))

    def test_create_mutated_model_duplication(self):
        # Given
        expected = CnnBaseArchitecture()
        expected = expected.build_from_file('resources/mutated_architecture_duplication.txt')

        architecture = CnnBaseArchitecture()
        architecture.build_from_file('resources/complex_architecture.txt')

        # When
        generator = ModelGenerator(architecture, 'keras', removal_rate=0, duplication_rate=1, amendment_rate=0, seed=14)
        actual = generator.create_mutated_model()
        mutated_arch = generator.get_last_mutation()

        # Then
        self.assertEqual(34, len(actual.model._layers))
        self.assertEqual(str(expected), str(mutated_arch))

    def test_create_mutated_model_amendments(self):
        # Given
        expected = CnnBaseArchitecture()
        expected = expected.build_from_file('resources/mutated_architecture_amendments.txt')

        architecture = CnnBaseArchitecture()
        architecture.build_from_file('resources/complex_architecture.txt')

        # When
        generator = ModelGenerator(architecture, 'keras', removal_rate=0, duplication_rate=0, amendment_rate=1, seed=13)
        actual = generator.create_mutated_model()
        mutated_arch = generator.get_last_mutation()

        # Then
        self.assertEqual(16, len(actual.model._layers))
        self.assertEqual(str(expected), str(mutated_arch))

    def test_create_mutated_model_removal(self):
        # Given
        expected = CnnBaseArchitecture()
        expected = expected.build_from_file('resources/mutated_architecture_removal.txt')

        architecture = CnnBaseArchitecture()
        architecture.build_from_file('resources/complex_architecture.txt')

        # When
        generator = ModelGenerator(architecture, 'keras', removal_rate=0.5, duplication_rate=0, amendment_rate=0, seed=13)
        actual = generator.create_mutated_model()
        mutated_arch = generator.get_last_mutation()

        # Then
        self.assertEqual(9, len(actual.model._layers))
        self.assertEqual(str(expected), str(mutated_arch))

    def test_unknown_builder(self):
        # Given
        architecture = CnnBaseArchitecture()

        # When
        generator = ModelGenerator(architecture, 'foo')

        # Then
        self.assertRaises(Exception, generator.create_base_model)

    def test_populate_builder_history(self):
        # Given
        expected = CnnBaseArchitecture()
        expected = expected.build_from_file('resources/mutated_architecture_history_duplicate.txt')

        architecture = CnnBaseArchitecture()
        architecture.build_from_file('resources/complex_architecture.txt')

        # When
        generator = ModelGenerator(architecture, 'keras', removal_rate=0, duplication_rate=0.8, amendment_rate=0, seed=14)
        generator.populate_history('resources/history', 'test-arch-', 3)
        actual = generator.create_mutated_model()
        mutated_arch = generator.get_last_mutation()

        # Then
        self.assertEqual(24, len(actual.model._layers))
        self.assertEqual(str(expected), str(mutated_arch))

    def test_get_last_empty_mutation(self):
        # Given
        expected = CnnBaseArchitecture()
        expected = expected.build_from_file('resources/complex_architecture.txt')

        architecture = CnnBaseArchitecture()
        architecture.build_from_file('resources/complex_architecture.txt')

        # When
        generator = ModelGenerator(architecture, 'keras')
        mutated_arch = generator.get_last_mutation()

        # Then
        self.assertEqual(str(expected), str(mutated_arch))